/* $Id: isaac_fit.c,v 1.4 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.4 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>

#include "isaac_fit.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup isaac_fit      Fitting
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief  Compute the non linearity of the detector
  @param  ilist   The list of exposures (images obtained with different DITs)
  @param  dit     The DIT values used
  @param  deg     The polynomial degree for the fit
  @note   dit _must_ point to (at least) as many doubles as there are images
          in the list, otherwise the behaviour of the function is undefined.
  @note   The created imagelist must be deallocated with cpl_imagelist_delete().
  @return The image list with the fit coefficients or NULL on error.
  
  Image list can be CPL_TYPE_FLOAT
  
  For each pixel, a polynomial representing the relation DIT=P(flux) is
  computed where:
  P(x) = a1*x + ... + an*x^n
  The returned image list contains deg+1 images.
  The deg first images contain the the deg first coefficients of the polynomial
  and the last image contains the error of the fit.
  
  Possible CPL error codes set in this function:
  - CPL_ERROR_NULL_INPUT if (one of) the input pointer(s) is NULL
  - CPL_ERROR_ILLEGAL_INPUT if deg is not strictly positive
  - CPL_ERROR_TYPE_MISMATCH if the passed image list type is not supported
  - CPL_ERROR_INCOMPATIBLE_INPUT if deg exceeds the number of input images
  - CPL_ERROR_SINGULAR_MATRIX if there are no pixels with enough distinct
    values, in this case a retry with a reduced deg may work.
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * isaac_imagelist_fit_polynomial(
        const cpl_imagelist *   ilist,
        const double        *   dit,
        int                     deg)
{
    cpl_imagelist   * fitres;
    const cpl_image * image;
    int               ni, nx, ny;

    double       f, f_prod;
    double      * pimad;
    const float * pi;
    cpl_matrix * mv;
    cpl_matrix * mb;
    cpl_matrix * mx;
    double     * pv;
    int          nsing = 0;
    int          i, j, k;

    /* Check entries */
    cpl_ensure(ilist != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(dit   != NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_ensure(deg > 0,       CPL_ERROR_ILLEGAL_INPUT, NULL);

    ni = cpl_imagelist_get_size(ilist);
    cpl_ensure(ni >= deg,     CPL_ERROR_INCOMPATIBLE_INPUT, NULL);

    image = cpl_imagelist_get_const(ilist, 0);

    nx = cpl_image_get_size_x(image);
    ny = cpl_image_get_size_y(image);

    /* Allocate deg+1 images to store the results */
    fitres = cpl_imagelist_new();
    for (i=0; i < deg+1; i++) {
        cpl_imagelist_set(fitres, cpl_image_new(nx, ny, CPL_TYPE_DOUBLE),
                          i);
    }

    /* Create matrices */
    pv = cpl_malloc(ni * deg * sizeof(double));
    mv = cpl_matrix_wrap(ni, deg, pv);

    /* Fill RHS matrix
       - mb is _not_ modified */
    mb = cpl_matrix_wrap(ni, 1, (double*)dit);

    /* Loop on pixels */
    for (i=0; i < nx*ny; i++) {
        /* Fill Vandermonde matrix */
        for (j=0; j < ni; j++) {
            pi=cpl_image_get_data_float_const(cpl_imagelist_get_const(ilist,j));
            f_prod = f = (double)pi[i];
            pv[deg * j] = f;
            for (k=1; k < deg; k++) {
                f_prod *= f;
                pv[deg * j + k] = f_prod;
            }
        }

        /* Solve least-squares */
        /* FIXME: See cpl_polynomial_fit_1d_create() for how to
           reduce the solve time */
        mx = cpl_matrix_solve_normal(mv, mb);

        if (mx == NULL) {
            for (k=0; k < deg+1; k++) {
                pimad =
                    cpl_image_get_data_double(cpl_imagelist_get(fitres,k));
                pimad[i] = 0.0;
            }
            nsing++;
        } else {
            double err, sq_err;
            const double * px = cpl_matrix_get_data(mx);

            /* Store results */
            for (k=0; k < deg; k++) {
                pimad = cpl_image_get_data_double(
                        cpl_imagelist_get(fitres, k));
                pimad[i] = px[k];
            }

            /* Goodness of fit */
            sq_err=0.0;
            for (j=0; j < ni; j++) {
                /* Error between model and reality */
                err = dit[j];
                /* Computed by model */
                for (k=0; k < deg; k++)
                    err -= px[k] * pv[deg * j + k];
                sq_err += err * err;
            }
            sq_err /= (double)ni;
            pimad=cpl_image_get_data_double(cpl_imagelist_get(fitres, deg));
            pimad[i] = sq_err;
            cpl_matrix_delete(mx);
        }
    }

    cpl_matrix_delete(mv);
    (void)cpl_matrix_unwrap(mb);

    if (nsing == nx*ny) {
        cpl_imagelist_delete(fitres);
        cpl_ensure(0, CPL_ERROR_SINGULAR_MATRIX, NULL);
    }

    return fitres;
}
#undef CPL_OPERATION

/**@}*/
