/* $Id: isaac_img_dark.c,v 1.38 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.38 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include <math.h>

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"

/*-----------------------------------------------------------------------------
                                Define
 -----------------------------------------------------------------------------*/

#define ISAAC_DARK_HSIZE_SW_DEF     6
#define ISAAC_DARK_HSIZE_LW_DEF     2

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_img_dark_create(cpl_plugin *);
static int isaac_img_dark_exec(cpl_plugin *);
static int isaac_img_dark_destroy(cpl_plugin *);
static int isaac_img_dark(cpl_parameterlist *, cpl_frameset *);

static int isaac_img_dark_avg_reduce(cpl_frameset *, cpl_image **);
static cpl_matrix * isaac_img_dark_ron_reduce(cpl_frameset *);
static int isaac_img_dark_compare(const cpl_frame *, const cpl_frame *); 
static int isaac_img_dark_save(cpl_image *, cpl_matrix *, int, cpl_frameset *, 
        cpl_parameterlist *, cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         hsize;
    int         nsamples;
    /* Outputs */
    double      dark_med;
    double      dark_stdev;
} isaac_img_dark_config;

static char isaac_img_dark_description[] = 
"isaac_img_dark -- ISAAC imaging dark recipe.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "ISAAC_IMG_DARK_RAW"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_img_dark",
                    "Dark recipe",
                    isaac_img_dark_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_img_dark_create,
                    isaac_img_dark_exec,
                    isaac_img_dark_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --nsamples */
    p = cpl_parameter_new_value("isaac.isaac_img_dark.nsamples",
            CPL_TYPE_INT, "number of samples for RON computation", 
            "isaac.isaac_img_dark", 100);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "nsamples");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);
    /* --hsize */
    p = cpl_parameter_new_value("isaac.isaac_img_dark.hsize",
            CPL_TYPE_INT, "half size of the window for RON computation", 
            "isaac.isaac_img_dark", -1);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "hsize");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_img_dark(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_parameter   *   par;
    cpl_size            nsets;
    cpl_size        *   selection;
    cpl_frameset    *   f_one;
    cpl_image       *   avg;
    cpl_matrix      *   rons;
    cpl_boolean         did_reduce = CPL_FALSE;
    int                 i;

    /* Retrieve input parameters */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_dark.hsize"); 
    isaac_img_dark_config.hsize = cpl_parameter_get_int(par);
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_dark.nsamples"); 
    isaac_img_dark_config.nsamples = cpl_parameter_get_int(par);

    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Labelize all input frames */
    cpl_msg_info(cpl_func, "Identify the different settings");
    selection=cpl_frameset_labelise(framelist, isaac_img_dark_compare, &nsets);
    if (selection == NULL) {
        cpl_msg_error(cpl_func, "Cannot labelise input frames");
        return CPL_ERROR_UNSPECIFIED;
    }
   
    /* Extract settings and reduce each of them */
    for (i=0; i<nsets; i++) {
        /* Initialise */
        isaac_img_dark_config.dark_med = -1.0;
        isaac_img_dark_config.dark_stdev = -1.0;
        avg = NULL;
        
        /* Reduce data set nb i */
        cpl_msg_info(cpl_func, "Reduce data set no %d out of %d", i+1,
                     (int)nsets);
        cpl_msg_indent_more();
        f_one = cpl_frameset_extract(framelist, selection, i);
        
        /* At least 2 frames required */
        if (cpl_frameset_get_size(f_one) < 2) {
            cpl_msg_warning(cpl_func, "Setting %d skipped (not enough frames)",
                    i+1);
        } else {
            /* AVG part */
            cpl_msg_info(cpl_func, "Compute the master dark");
            cpl_msg_indent_more();
            if (isaac_img_dark_avg_reduce(f_one, &avg)) {
                cpl_msg_warning(cpl_func, "Cannot reduce set number %d", i+1);
            }
            cpl_msg_indent_less();
            /* RON part */
            cpl_msg_info(cpl_func, "Compute the read-out noise");
            cpl_msg_indent_more();
            if ((rons = isaac_img_dark_ron_reduce(f_one)) == NULL) {
                cpl_msg_warning(cpl_func, "Cannot reduce set number %d", i+1);
            }
            cpl_msg_indent_less();
            /* Save the products */
            if (isaac_img_dark_save(avg, rons, i+1, f_one, parlist, 
                        framelist) !=0 ){
                cpl_msg_error(cpl_func, "Cannot save the products");
                if (avg) cpl_image_delete(avg);
                if (rons) cpl_matrix_delete(rons);
                return CPL_ERROR_UNSPECIFIED;
            }
            if (rons) cpl_matrix_delete(rons);
            if (!cpl_error_get_code()) did_reduce = CPL_TRUE;
        }
        if (avg) cpl_image_delete(avg);
        cpl_frameset_delete(f_one);
        cpl_msg_indent_less();
    }
    
    /* Free and return */
    cpl_free(selection); 

    cpl_ensure_code(did_reduce, CPL_ERROR_ILLEGAL_INPUT);

    return cpl_error_set_where(cpl_func); /* Propagate error, if any */
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction ('avg' part) is implemented here 
  @param    framelist   the frames list of the current set 
  @param    avg         the averaged image
  @return   0 if ok, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_avg_reduce(
        cpl_frameset    *   framelist,
        cpl_image       **  avg)
{
    cpl_imagelist   *   iset;
    cpl_vector      *   medians;
    cpl_image       *   image;
    int                 i;
                    
    /* Test entries */
    if (framelist == NULL) return CPL_ERROR_UNSPECIFIED;
    
    /* Load the image set */
    if ((iset = cpl_imagelist_load_frameset(framelist, CPL_TYPE_FLOAT, 1, 
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the data");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Average it to the master dark */
    if ((*avg = cpl_imagelist_collapse_create(iset)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot average the data set");
        cpl_imagelist_delete(iset);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Compute mean-rms of the median values */
    medians = cpl_vector_new(cpl_imagelist_get_size(iset));
    for (i=0; i<cpl_imagelist_get_size(iset); i++) {
        image = cpl_imagelist_get(iset, i);
        cpl_vector_set(medians, i, cpl_image_get_median(image));
    }
    cpl_imagelist_delete(iset);
    isaac_img_dark_config.dark_med = cpl_vector_get_mean(medians);
    isaac_img_dark_config.dark_stdev = cpl_vector_get_stdev(medians);

    /* Free and Return */
    cpl_vector_delete(medians);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction ('ron' part) is implemented here 
  @param    framelist   the frames list of the current set 
  @param    set_nb      the setting number
  @param    set         the complete frames set
  @return   the matrix with the RONs values

  The return matrix has either 1 (LW) or 4 (SW) columns.
  In SW mode the 4 columns are for the ll, lr, ul, ur quadrants.
 */
/*----------------------------------------------------------------------------*/
static cpl_matrix * isaac_img_dark_ron_reduce(cpl_frameset * framelist)
{
    cpl_frame           *   cur_frame;
    cpl_propertylist    *   plist;
    char                    arm;
    const char          *   sval;
    cpl_imagelist       *   iset;
    cpl_matrix          *   rons;
    cpl_image           *   tmp_im;
    double                  rms;
    double                  norm;
    int                     ndit;
    cpl_size                zone_def[4];
    int                     i;

    /* Test entries */
    if (framelist == NULL) return NULL;

    /* Initialise */
    rons = NULL;
    
    /* Check the arm used */
    if (cpl_error_get_code()) return NULL;
    cur_frame = cpl_frameset_get_position(framelist, 0);
    plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame), 0);
    sval = isaac_pfits_get_arm(plist);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "Cannot get the used arm");
        cpl_propertylist_delete(plist);
        return NULL;
    }
    arm = (int)(sval[0]);
    cpl_propertylist_delete(plist);

    /* Load the current set */
    if ((iset = cpl_imagelist_load_frameset(framelist, CPL_TYPE_FLOAT, 1, 
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot load the data");
        return NULL;
    }
   
    /* Switch on the arm */
    switch (arm) {
        case 'S':
            /* Create the matrix */
            rons = cpl_matrix_new(cpl_imagelist_get_size(iset)-1, 4);
            if (isaac_img_dark_config.hsize < 0) 
                isaac_img_dark_config.hsize = ISAAC_DARK_HSIZE_SW_DEF;

            /* Loop on all pairs */
            for (i=0; i<cpl_imagelist_get_size(iset)-1; i++) {
                cpl_msg_info(cpl_func, "Pair number %d", i+1);
                /* Get the norm factor */
                if (cpl_error_get_code()) {
                    cpl_matrix_delete(rons);
                    cpl_imagelist_delete(iset);
                    return NULL;
                }
                cur_frame = cpl_frameset_get_position(framelist, i);
                plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
                        0);
                ndit = isaac_pfits_get_ndit(plist);
                cpl_propertylist_delete(plist);
                if (cpl_error_get_code()) {
                    cpl_msg_error(cpl_func, "Cannot get the NDIT");
                    cpl_matrix_delete(rons);
                    cpl_imagelist_delete(iset);
                    return NULL;
                }
                norm = 0.5 * ndit;
                norm = sqrt(norm);
                
                /* Compute the current subtracted image */
                if ((tmp_im = cpl_image_subtract_create(
                                cpl_imagelist_get(iset, i),
                                cpl_imagelist_get(iset, i+1))) == NULL) {
                    cpl_msg_error(cpl_func, "Cannot subtract the images");
                    cpl_imagelist_delete(iset);
                    cpl_matrix_delete(rons);
                    return NULL;
                }

                /* Get measurement for lower-left quadrant */
                zone_def[0] = 1;
                zone_def[1] = cpl_image_get_size_x(tmp_im)/2;
                zone_def[2] = 1;
                zone_def[3] = cpl_image_get_size_y(tmp_im)/2;
                cpl_flux_get_noise_window(tmp_im, zone_def, 
                        isaac_img_dark_config.hsize, 
                        isaac_img_dark_config.nsamples, &rms, NULL);
                cpl_matrix_set(rons, i, 0, rms * norm);
                cpl_msg_info(cpl_func, "RON in LL quadrant: %g", rms * norm);
 
                /* Get measurement for lower-right quadrant */
                zone_def[0] = cpl_image_get_size_x(tmp_im)/2 + 1;
                zone_def[1] = cpl_image_get_size_x(tmp_im);
                zone_def[2] = 1;
                zone_def[3] = cpl_image_get_size_y(tmp_im)/2;
                cpl_flux_get_noise_window(tmp_im, zone_def, 
                        isaac_img_dark_config.hsize, 
                        isaac_img_dark_config.nsamples, &rms, NULL);
                cpl_matrix_set(rons, i, 1, rms * norm);
                cpl_msg_info(cpl_func, "RON in LR quadrant: %g", rms * norm);
                
                /* Get measurement for upper-left quadrant */
                zone_def[0] = 1;
                zone_def[1] = cpl_image_get_size_x(tmp_im)/2;
                zone_def[2] = cpl_image_get_size_y(tmp_im)/2 + 1;
                zone_def[3] = cpl_image_get_size_y(tmp_im);
                cpl_flux_get_noise_window(tmp_im, zone_def, 
                        isaac_img_dark_config.hsize, 
                        isaac_img_dark_config.nsamples, &rms, NULL);
                cpl_matrix_set(rons, i, 2, rms * norm);
                cpl_msg_info(cpl_func, "RON in UL quadrant: %g", rms * norm);
                
                /* Get measurement for upper-right quadrant */
                zone_def[0] = cpl_image_get_size_x(tmp_im)/2 + 1;
                zone_def[1] = cpl_image_get_size_x(tmp_im);
                zone_def[2] = cpl_image_get_size_y(tmp_im)/2 + 1;
                zone_def[3] = cpl_image_get_size_y(tmp_im);
                cpl_flux_get_noise_window(tmp_im, zone_def, 
                        isaac_img_dark_config.hsize, 
                        isaac_img_dark_config.nsamples, &rms, NULL);
                cpl_matrix_set(rons, i, 3, rms * norm);
                cpl_msg_info(cpl_func, "RON in UR quadrant: %g", rms * norm);
                cpl_image_delete(tmp_im);
            }
            break;
        case 'L':
            /* Create the matrix */
            rons = cpl_matrix_new(cpl_imagelist_get_size(iset)-1, 1);
            if (isaac_img_dark_config.hsize < 0) 
                isaac_img_dark_config.hsize = ISAAC_DARK_HSIZE_LW_DEF;

            /* Loop on all pairs */
            for (i=0; i<cpl_imagelist_get_size(iset)-1; i++) {
                cpl_msg_info(cpl_func, "Pair number %d", i+1);
                /* Get the norm factor */
                if (cpl_error_get_code()) {
                    cpl_matrix_delete(rons);
                    cpl_imagelist_delete(iset);
                    return NULL;
                }
                cur_frame = cpl_frameset_get_position(framelist, i);
                plist = cpl_propertylist_load(cpl_frame_get_filename(cur_frame),
                        0);
                ndit = isaac_pfits_get_ndit(plist);
                cpl_propertylist_delete(plist);
                if (cpl_error_get_code()) {
                    cpl_msg_error(cpl_func, "Cannot get the NDIT");
                    cpl_matrix_delete(rons);
                    cpl_imagelist_delete(iset);
                    return NULL;
                }
                norm = 0.5 * ndit;
                norm = sqrt(norm);
                
                /* Compute the current subtracted image */
                if ((tmp_im = cpl_image_subtract_create(
                                cpl_imagelist_get(iset, i),
                                cpl_imagelist_get(iset, i+1))) == NULL) {
                    cpl_msg_error(cpl_func, "Cannot subtract the images");
                    cpl_imagelist_delete(iset);
                    cpl_matrix_delete(rons);
                    return NULL;
                }

                /* Get measurement */
                cpl_flux_get_noise_window(tmp_im, NULL, 
                        isaac_img_dark_config.hsize, 
                        isaac_img_dark_config.nsamples, &rms, NULL);
                cpl_matrix_set(rons, i, 0, rms * norm);
                cpl_msg_info(cpl_func, "RON: %g", rms * norm);
                cpl_image_delete(tmp_im);
            }
            break;
        default:
            cpl_msg_error(cpl_func, "Unsupported arm");
            cpl_matrix_delete(rons);
            rons = NULL;
            break;
    }
    
    /* Free and return */
    cpl_imagelist_delete(iset);
    return rons;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the dark recipe products on disk
  @param    avg     the average image
  @param    rons    the RON values
  @param    set_nb  the current setting number
  @param    set      the frame set for the current setting
  @param    parlist     the input list of parameters
  @param    set_tot the total input frame set to register the products
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_save(
        cpl_image           *   avg,
        cpl_matrix          *   rons,
        int                     set_nb,
        cpl_frameset        *   set,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set_tot)
{
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    const cpl_frame           *   ref_frame;
    char                *   filename;
    char                    qc_str[128];
    int                     i;

    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
    cpl_propertylist_append_double(qclist, "ESO QC DARKMED",
            isaac_img_dark_config.dark_med);
    cpl_propertylist_append_double(qclist, "ESO QC DARKSTDEV",
            isaac_img_dark_config.dark_stdev);
    if (rons != NULL) {
        switch (cpl_matrix_get_ncol(rons)) {
            case 1:
                for (i=0; i<cpl_matrix_get_nrow(rons); i++) {
                    sprintf(qc_str, "ESO QC RON%d", i+1);
                    cpl_propertylist_append_double(qclist, qc_str, 
                            cpl_matrix_get(rons, i, 0));
                }
                break;
            case 4:
                for (i=0; i<cpl_matrix_get_nrow(rons); i++) {
                    sprintf(qc_str, "ESO QC LL RON%d", i+1);
                    cpl_propertylist_append_double(qclist, qc_str, 
                            cpl_matrix_get(rons, i, 0));
                    sprintf(qc_str, "ESO QC LR RON%d", i+1);
                    cpl_propertylist_append_double(qclist, qc_str, 
                            cpl_matrix_get(rons, i, 1));
                    sprintf(qc_str, "ESO QC UL RON%d", i+1);
                    cpl_propertylist_append_double(qclist, qc_str, 
                            cpl_matrix_get(rons, i, 2));
                    sprintf(qc_str, "ESO QC UR RON%d", i+1);
                    cpl_propertylist_append_double(qclist, qc_str, 
                            cpl_matrix_get(rons, i, 3));
                }
                break;
            default:
                cpl_msg_error(cpl_func, "Invalid RONs matrix format");
                break;
        }
    }
    
    /* Write the average image */
    filename = cpl_sprintf("isaac_img_dark_set%02d_avg.fits", set_nb);
    irplib_dfs_save_image(set_tot,
            parlist,
            set,
            avg,
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_dark",
            ISAAC_IMG_DARK_AVG,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            filename);
    cpl_free(filename);

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);
    
    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }
 
    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|ESO TPL ID|ESO DPR TECH|DATE-OBS|ESO DET DIT|"
        "ESO DET NDIT|ESO DET NCORRS|ESO DET NDSAMPLES|ESO DET MODE NAME)$", 0);
    cpl_propertylist_delete(plist);
   
    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_DARK_AVG);

    /* Save the PAF file */
    filename = cpl_sprintf("isaac_img_dark_set%02d.paf", set_nb);
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_dark",
            paflist,
            filename);
    cpl_free(filename);
    cpl_propertylist_delete(paflist);
    return  0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Comparison function to identify different settings
  @param    frame1  first frame 
  @param    frame2  second frame 
  @return   0 if frame1!=frame2, 1 if frame1==frame2, -1 in error case
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_dark_compare(
        const cpl_frame *   frame1, 
        const cpl_frame *   frame2) 
{
    int                     comparison;
    cpl_propertylist    *   plist1;
    cpl_propertylist    *   plist2;
    double                  dval1, dval2;
    int                     ival1, ival2;

    /* Test entries */
    if (frame1==NULL || frame2==NULL) return CPL_ERROR_UNSPECIFIED;

    /* Get property lists */
    if ((plist1=cpl_propertylist_load(cpl_frame_get_filename(frame1),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((plist2=cpl_propertylist_load(cpl_frame_get_filename(frame2),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(plist1);
        return CPL_ERROR_UNSPECIFIED;
    }
    
    /* Test status */
    if (cpl_error_get_code()) {
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
     
    /* Compare exposure time */
    comparison = 1;
    dval1 = isaac_pfits_get_dit(plist1);
    dval2 = isaac_pfits_get_dit(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get exposure time");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (fabs(dval1-dval2) > 1e-5) comparison = 0;

    /* Compare NDIT */
    ival1 = isaac_pfits_get_ndit(plist1);
    ival2 = isaac_pfits_get_ndit(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get NDIT");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (ival1 != ival2) comparison = 0;

    /* Compare the readout mode */
    ival1 = isaac_pfits_get_rom(plist1);
    ival2 = isaac_pfits_get_rom(plist2);
    if (cpl_error_get_code()) {
        cpl_msg_error(cpl_func, "cannot get read-out mode");
        cpl_propertylist_delete(plist1);
        cpl_propertylist_delete(plist2);
        return CPL_ERROR_UNSPECIFIED;
    }
    if (ival1 != ival2) comparison = 0;

    /* Free and return */
    cpl_propertylist_delete(plist1);
    cpl_propertylist_delete(plist2);
    return comparison;
}

