/* $Id: isaac_img_detlin.c,v 1.31 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.31 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"

#include "isaac_utils.h"
#include "isaac_pfits.h"
#include "isaac_dfs.h"
#include "isaac_fit.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_img_detlin_create(cpl_plugin *);
static int isaac_img_detlin_exec(cpl_plugin *);
static int isaac_img_detlin_destroy(cpl_plugin *);
static int isaac_img_detlin(cpl_parameterlist *, cpl_frameset *);

static cpl_imagelist * isaac_img_detlin_load(cpl_frameset *, cpl_frameset *, 
        double **);
static int isaac_img_detlin_save(cpl_imagelist *, cpl_parameterlist *, 
        cpl_frameset *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int     force_flag;

    /* Outputs */
    double  lamp_stability;
} isaac_img_detlin_config;

static char isaac_img_detlin_description[] =
"isaac_img_detlin -- ISAAC imaging detector linearity recipe.\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "ISAAC_IMG_DETLIN_LAMP_RAW" or\n"
"raw-file.fits "ISAAC_IMG_DETLIN_DARK_RAW"\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_img_detlin",
                    "Detector linearity recipe",
                    isaac_img_detlin_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_img_detlin_create,
                    isaac_img_detlin_exec,
                    isaac_img_detlin_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_detlin_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new(); 

    /* Fill the parameters list */
    /* --force */
    p = cpl_parameter_new_value("isaac.isaac_img_detlin.force", CPL_TYPE_BOOL,
            "flag to force th computation", "isaac.isaac_img_detlin",
            FALSE);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "force");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_detlin_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_img_detlin(recipe->parameters, recipe->frames);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_detlin_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Get the command line options and execute the data reduction
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_detlin(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   framelist)
{
    cpl_parameter   *   par;
    cpl_frameset    *   darkframes;
    cpl_frameset    *   lampframes;
    cpl_imagelist   *   iset;
    double          *   ditval;
    cpl_imagelist   *   fitres;

    /* Initialise */
    par = NULL;

    /* Retrieve input parameters */
    /* Force flag */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_img_detlin.force");
    isaac_img_detlin_config.force_flag = cpl_parameter_get_bool(par);

    /* Identify the RAW and CALIB frames in the input frameset */
    if (isaac_dfs_set_groups(framelist)) {
        cpl_msg_error(cpl_func, "Cannot identify RAW and CALIB frames");
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Retrieve raw frames */
    if ((lampframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_DETLIN_LAMP_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find any lamp frame in input");
        return CPL_ERROR_UNSPECIFIED;
    }
    if ((darkframes = isaac_extract_frameset(framelist,
                    ISAAC_IMG_DETLIN_DARK_RAW)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot find any dark frame in input");
        cpl_frameset_delete(lampframes);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Load the data and check the DIT consistency */
    cpl_msg_info(cpl_func, "Load the data");
    cpl_msg_indent_more();
    if ((iset = isaac_img_detlin_load(lampframes, darkframes, &ditval))==NULL) {
        cpl_msg_error(cpl_func, "Cannot load the data");
        cpl_frameset_delete(lampframes);
        cpl_frameset_delete(darkframes);
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_frameset_delete(lampframes);
    cpl_frameset_delete(darkframes);
    cpl_msg_indent_less();
   
    /* Compute the linearity coefficients */
    cpl_msg_info(cpl_func, "Compute the linearity coefficients");
    cpl_msg_indent_more();
    if ((fitres = isaac_imagelist_fit_polynomial(iset, ditval, 3)) == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute the linearity coefficients");
        cpl_imagelist_delete(iset);
        cpl_free(ditval);
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_indent_less();
    cpl_free(ditval);
    cpl_imagelist_delete(iset);

    /* Save the products */
    cpl_msg_info(cpl_func, "Save the products");
    cpl_msg_indent_more();
    if (isaac_img_detlin_save(fitres, parlist, framelist)==-1) {
        cpl_msg_error(cpl_func, "Cannot save the products");
        cpl_imagelist_delete(fitres);
        cpl_msg_indent_less();
        return CPL_ERROR_UNSPECIFIED;
    }
    cpl_msg_indent_less();

    /* Free and return */
    cpl_imagelist_delete(fitres);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Load the data and check their consistency
  @param    lamps   the lamp frames
  @param    darks   the dark frames
  @param    ditval  the DIT values
  @return   the set of images 
 */
/*----------------------------------------------------------------------------*/
static cpl_imagelist * isaac_img_detlin_load(
        cpl_frameset    *   lamps,
        cpl_frameset    *   darks,
        double          **  ditvals)
{
    int                     nb_lamps;
    cpl_vector          *   selection;
    cpl_frame           *   frame;
    cpl_propertylist    *   propertylist;
    double                  dit_lamp, dit_dark;
    int                     dit_stab;
    cpl_imagelist       *   lamps_data;
    cpl_imagelist       *   darks_data;
    double              *   stab_levels;
    double              *   dit_purged;
    int                     i, j;

    /* Check that there are as many lamp as darks */
    if ((nb_lamps = cpl_frameset_get_size(lamps)) < 3) return NULL;
    if (cpl_frameset_get_size(darks) != nb_lamps) return NULL;

    /* Check out that they have consistent integration times */
    cpl_msg_info(cpl_func, "Checking DIT consistency");
    selection = cpl_vector_new(nb_lamps);
    *ditvals = cpl_malloc(nb_lamps * sizeof(double));
    dit_stab = 0;
    for (i=0; i<nb_lamps; i++) {
        /* Check if ok */
        if (cpl_error_get_code()) {
            cpl_vector_delete(selection);
            cpl_free(*ditvals);
            return NULL;
        }
        /* DIT from LAMP */
        frame = cpl_frameset_get_position(lamps, i);
        propertylist=cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
        dit_lamp = (double)isaac_pfits_get_dit(propertylist);
        cpl_propertylist_delete(propertylist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "Cannot get DIT");
            cpl_vector_delete(selection);
            cpl_free(*ditvals);
            return NULL;
        }
        /* DIT from DARK */
        frame = cpl_frameset_get_position(darks, i);
        propertylist=cpl_propertylist_load(cpl_frame_get_filename(frame), 0);
        dit_dark = (double)isaac_pfits_get_dit(propertylist);
        cpl_propertylist_delete(propertylist);
        if (cpl_error_get_code()) {
            cpl_msg_error(cpl_func, "Cannot get DIT");
            cpl_vector_delete(selection);
            cpl_free(*ditvals);
            return NULL;
        }
        /* Check consistency */
        if (fabs(dit_dark-dit_lamp) > 1e-3) {
            cpl_msg_error(cpl_func, "DIT not consistent between LAMP and DARK");
            cpl_vector_delete(selection);
            cpl_free(*ditvals);
            return NULL;
        }
        (*ditvals)[i] = dit_lamp;
        /* Set selection */
        if (i==0) {
            cpl_vector_set(selection, i, -1.0);
            dit_stab ++;
        } else {
            if (fabs(dit_lamp - (*ditvals)[0]) < 1e-5) {
                cpl_vector_set(selection, i, -1.0);
                dit_stab ++;
            } else {
                cpl_vector_set(selection, i, 1.0);
            }
        }
    }
    
    /* Check if there are enough DITs for stability check */
    if (dit_stab < 2) {
        cpl_msg_error(cpl_func, "Not enough frames for stability check");
        cpl_vector_delete(selection);
        cpl_free(*ditvals);
        return NULL;
    }

    /* Load the data and compute lamp-dark */
    cpl_msg_info(cpl_func, "Compute the differences lamp - dark");
    lamps_data = cpl_imagelist_load_frameset(lamps, CPL_TYPE_FLOAT, 1, 0);
    darks_data = cpl_imagelist_load_frameset(darks, CPL_TYPE_FLOAT, 1, 0);
    if (cpl_imagelist_subtract(lamps_data,darks_data) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "Cannot subtract the 2 image lists");
        cpl_vector_delete(selection);
        cpl_free(*ditvals);
        if (lamps_data) cpl_imagelist_delete(lamps_data);
        if (darks_data) cpl_imagelist_delete(darks_data);
        return NULL;
    }
    cpl_imagelist_delete(darks_data);
    
    /* Check the lamp stability */
    cpl_msg_info(cpl_func, "Check the lamp stability");
    stab_levels = cpl_malloc(dit_stab * sizeof(double));
    j = 0;
    for (i=0; i<nb_lamps; i++) {
        if (cpl_vector_get(selection, i) < 0) {
            stab_levels[j] = 
                cpl_image_get_mean(cpl_imagelist_get(lamps_data, i));
            j++;
        }
    }

    /* Compute the lamp stability */
    isaac_img_detlin_config.lamp_stability = 0.0;
    for (i=1; i<dit_stab; i++) {
        if ((fabs(stab_levels[i]-stab_levels[0]) / stab_levels[0]) >
                isaac_img_detlin_config.lamp_stability) 
            isaac_img_detlin_config.lamp_stability = 
                fabs(stab_levels[i]-stab_levels[0]) / stab_levels[0];
    }
    cpl_free(stab_levels);
    
    /* Check the lamp stability */
    if (isaac_img_detlin_config.lamp_stability > 0.01) {
        if (isaac_img_detlin_config.force_flag == 1) {
            cpl_msg_warning(cpl_func, 
                    "level difference #%d too high - proceed anyway",i+1);
        } else {
            cpl_msg_error(cpl_func, "level difference #%d too high", i+1);
            cpl_vector_delete(selection);
            cpl_free(*ditvals);
            cpl_imagelist_delete(lamps_data);
            return NULL;
        }
    }

    /* Purge the stability check data and DITs */
    if (cpl_imagelist_erase(lamps_data, selection) != CPL_ERROR_NONE) {
        cpl_msg_error(cpl_func, "cannot discard stability frames");
        cpl_vector_delete(selection);
        cpl_free(*ditvals);
        cpl_imagelist_delete(lamps_data);
        return NULL;
    }
    dit_purged = cpl_malloc(cpl_imagelist_get_size(lamps_data)*sizeof(double));
    j = 0;
    for (i=0; i<nb_lamps; i++) {
        if (cpl_vector_get(selection, i) > 0) {
            dit_purged[j] = (*ditvals)[i];
            j++;
        }
    }
    cpl_free(*ditvals);
    *ditvals = dit_purged;
    
    /* Free and eturn  */
    cpl_vector_delete(selection);
    return lamps_data;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the different products of the recipe
  @param    fitres      the 3 fit coefficients images and the fit quality
  @param    parlist     the input list of parameters
  @param    set         the input frame set
  @return   0 if everything is ok, -1 otherwise
 */
/*----------------------------------------------------------------------------*/
static int isaac_img_detlin_save(
        cpl_imagelist       *   fitres,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    const cpl_frame           *   ref_frame;
    cpl_propertylist    *   plist;
    cpl_propertylist    *   qclist;
    cpl_propertylist    *   paflist;
    double                  qc_meda, qc_medb, qc_medc, qc_medq;

    /* Compute the QC params */
    qc_meda = cpl_image_get_median(cpl_imagelist_get(fitres, 0));
    qc_medb = cpl_image_get_median(cpl_imagelist_get(fitres, 1));
    qc_medc = cpl_image_get_median(cpl_imagelist_get(fitres, 2));
    qc_medq = cpl_image_get_median(cpl_imagelist_get(fitres, 3));
   
    /* Get the QC params in qclist */
    qclist = cpl_propertylist_new();
    cpl_propertylist_append_double(qclist, "ESO QC DETLIN MEDA", qc_meda);
    cpl_propertylist_append_double(qclist, "ESO QC DETLIN MEDB", qc_medb);
    cpl_propertylist_append_double(qclist, "ESO QC DETLIN MEDC", qc_medc);
    cpl_propertylist_append_double(qclist, "ESO QC DETLIN MEDQ", qc_medq);
    cpl_propertylist_append_double(qclist, "ESO QC DETLIN LAMP", 
            isaac_img_detlin_config.lamp_stability );

    /* Write the _A FITS FILE */
    irplib_dfs_save_image(set,
            parlist,
            set,
            cpl_imagelist_get(fitres, 0),
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_detlin",
            ISAAC_IMG_DETLIN_A,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_detlin_A.fits");

    /* Write the _B FITS FILE */
    irplib_dfs_save_image(set,
            parlist,
            set,
            cpl_imagelist_get(fitres, 1),
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_detlin",
            ISAAC_IMG_DETLIN_B,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_detlin_B.fits");

    /* Write the _C FITS FILE */
    irplib_dfs_save_image(set,
            parlist,
            set,
            cpl_imagelist_get(fitres, 2),
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_detlin",
            ISAAC_IMG_DETLIN_C,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_detlin_C.fits");

    /* Write the _Q FITS FILE */
    irplib_dfs_save_image(set,
            parlist,
            set,
            cpl_imagelist_get(fitres, 3),
            CPL_BPP_IEEE_FLOAT,
            "isaac_img_detlin",
            ISAAC_IMG_DETLIN_Q,
            qclist,
            NULL,
            PACKAGE "/" PACKAGE_VERSION,
            "isaac_img_detlin_Q.fits");

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW);

    /* Get FITS header from reference file */
    if ((plist=cpl_propertylist_load(cpl_frame_get_filename(ref_frame),
                    0)) == NULL) {
        cpl_msg_error(cpl_func, "getting header from reference frame");
        cpl_propertylist_delete(qclist);
        return CPL_ERROR_UNSPECIFIED;
    }

    /* Get the keywords for the paf file */
    paflist = cpl_propertylist_new();
    cpl_propertylist_copy_property_regexp(paflist, plist,
        "^(ARCFILE|MJD-OBS|ESO TPL ID|DATE-OBS|ESO DET DIT|"
        "ESO DET NDIT|ESO DET NCORRS|ESO DET MODE NAME)$", 0);
    cpl_propertylist_delete(plist);

    /* Copy the QC in paflist */
    cpl_propertylist_copy_property_regexp(paflist, qclist, ".", 0);
    cpl_propertylist_delete(qclist);

    /* PRO.CATG */
    cpl_propertylist_update_string(paflist, CPL_DFS_PRO_CATG,
                                   ISAAC_IMG_DETLIN_A);

    /* Save the PAF file */
    cpl_dfs_save_paf("ISAAC",
            "isaac_img_detlin",
            paflist,
            "isaac_img_detlin_QC.paf");
    cpl_propertylist_delete(paflist);
    return  0;
}

