/* $Id: isaac_spc_wlmodel.c,v 1.7 2013-03-12 08:06:48 llundin Exp $
 *
 * This file is part of the ISAAC Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2013-03-12 08:06:48 $
 * $Revision: 1.7 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "isaac_physicalmodel.h"
#include "isaac_utils.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int isaac_spc_wlmodel_create(cpl_plugin *);
static int isaac_spc_wlmodel_exec(cpl_plugin *);
static int isaac_spc_wlmodel_destroy(cpl_plugin *);
static int isaac_spc_wlmodel(cpl_parameterlist *);
 
/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char isaac_spc_wlmodel_description[] = 
"isaac_spc_wlmodel -- ISAAC spectro wavelength model\n";

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok

  This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof(*recipe));
    cpl_plugin  *   plugin = &recipe->interface;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    ISAAC_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "isaac_spc_wlmodel",
                    "Spectro wavelength model recipe",
                    isaac_spc_wlmodel_description,
                    "Lars Lundin",
                    PACKAGE_BUGREPORT,
                    isaac_get_license(),
                    isaac_spc_wlmodel_create,
                    isaac_spc_wlmodel_exec,
                    isaac_spc_wlmodel_destroy);

    cpl_pluginlist_append(list, plugin);
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlmodel_create(cpl_plugin * plugin)
{
    cpl_recipe      * recipe;
    cpl_parameter   * p;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new();

    /* Fill the parameters list */
    /* --wcen */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlmodel.wcen",CPL_TYPE_DOUBLE,
            "the central wavelength in angstroms", "isaac.isaac_spc_wlmodel", 
            22000.0);
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "wcen");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --obj */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlmodel.obj",
            CPL_TYPE_STRING, "objective (S1/S2/L1/L2/L3)", 
            "isaac.isaac_spc_wlmodel", "S2");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "obj");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    /* --res */
    p = cpl_parameter_new_value("isaac.isaac_spc_wlmodel.res",
            CPL_TYPE_STRING, "resolution (LR/MR)", 
            "isaac.isaac_spc_wlmodel", "MR");
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "res");
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, p);

    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlmodel_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    return isaac_spc_wlmodel(recipe->parameters);
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlmodel_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe;

    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE)
        recipe = (cpl_recipe *)plugin;
    else return CPL_ERROR_UNSPECIFIED;

    cpl_parameterlist_delete(recipe->parameters);
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    parlist     the parameters list
  @param    framelist   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int isaac_spc_wlmodel(cpl_parameterlist * parlist)
{
    cpl_parameter       *   par;
    double              *   wlcal;
    double                  wcen;
    const char          *   obj;
    const char          *   res;
    FILE                *   fp;
    int                     i;
    
    /* Initialise */
    par = NULL;

    /* Retrieve input parameters */
    /* --wcen */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlmodel.wcen");
    wcen = cpl_parameter_get_double(par);

    /* --obj */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlmodel.obj");
    obj = cpl_parameter_get_string(par);

    /* --res */
    par = cpl_parameterlist_find(parlist, "isaac.isaac_spc_wlmodel.res");
    res = cpl_parameter_get_string(par);

    /* Physical model */
    wlcal = isaac_physical_model(wcen, obj, res, 1024);
    if (wlcal == NULL) {
        cpl_msg_error(cpl_func, "Cannot compute the physical model");
        return CPL_ERROR_UNSPECIFIED;
    } 
            
    /* Save results */
    if ((fp = fopen("wavecal.txt", "w")) == NULL) {
        cpl_msg_error(cpl_func, "Cannot open the output file");
        cpl_free(wlcal);
        return CPL_ERROR_UNSPECIFIED;
    } 
    for (i=0; i<1024; i++) fprintf(fp, "%d\t\t%g\n", i+1, wlcal[i]);
    fclose(fp);
    cpl_free(wlcal);
    return 0;
}

